/*  Copyright 2007 Linux Box, Inc.  Ryan Hughes <ryan@linuxbox.com>.
 *  Licensed under the Apache License, Version 2.0 (the "License"); you may not
 *  use this file except in compliance with the License. You may obtain a copy
 *  of the License at 
 *          http://www.apache.org/licenses/LICENSE-2.0 
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *  License for the specific language governing permissions and limitations
 *  under the License.
 */

/**
 * Provide timeouts for auth modules.  (Compatible with apache 2.2).
 * Enable the module by adding something like this to your httpd.conf:
 * See README.txt for info on how to use it.
 */

#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_main.h"
#include "http_protocol.h"
#include "http_request.h"
#include "util_script.h"
#include "http_connection.h"

#include "apr_base64.h"
#include "apr_dbm.h"
#include "apr_errno.h"
#include "apr_file_io.h"
#include "apr_md5.h"
#include "apr_shm.h"
#include "apr_strings.h"

#include <ctype.h>
#include <stdio.h>
#include <time.h>
#include <unistd.h>

#include <sys/types.h>

/*--------------------------------------------------------------------------*/
/*                                                                          */
/* Data declarations.                                                       */
/*                                                                          */
/* Here are the static cells and structure declarations private to our      */
/* module.                                                                  */
/*                                                                          */
/*--------------------------------------------------------------------------*/

/**
 * Configuration structure.
 * Module-specific struct to store configuration options.
 */
typedef struct x_cfg {
    apr_pool_t *pool;

    int cmode;                  /* Environment to which record applies
                                 * (directory, server, or combination).
                                 */
#define CONFIG_MODE_SERVER 1
#define CONFIG_MODE_DIRECTORY 2
#define CONFIG_MODE_COMBO 3     /* Shouldn't ever happen. */

    char *loc;                  // Where does this config apply to?

    int timeout_time;           // How many seconds does it take to time out
    char *file_for_db;          // What file is the db kept in?
    int clean_howoften;         // After how many runs shall we clean the db?
    int db_persists;            // Does the db persist across server restarts?
	int helpful_ui;				// Explain why you hafta log in twice?
    char *anon_url;             // The template to give to anonymous users.
    char *nonanon_url;          // The template to give to non-anonymous users.
    char *logout_url;			// URL that causes you to expire immediately.
    char *logout_redirect_default; // URL you are redirected to if you don't
                                    // have a referrer set.
    char *logout_redirect_force; // URL you are redirected to *even if* you
                                    // have a referrer set.
    char *alreadyloggedout_url; // URL for if you're already logged out.
    apr_table_t *cookies;       // The cookies, as a table with key/val.
} x_cfg;

/*------------------------------------------------------*
 * Configuration Defaults
 *------------------------------------------------------*/

// The default of AuthTimeOut
#define ATO_TIMEOUT 0

// The default of AuthTimeOutHelpfulUI
#define ATO_HELPFULUI 1

// The default of AuthTimeOutCleanPeriod
#define ATO_CLEAN_HOWOFTEN 10000

// A portion of the default dbm filename.
#define ATO_DBM_FILENAME "authtimeout_db"

/**
 * Get the default db filename.
 * This will allocate a new string, with the default dbm filename.  
 * The default is <tempdir>/<ATO_DBM_FILENAME>.
 * The new string is allocated in the context of the pool you give.
 *
 * @param p The pool in which to allocate the string we're returning to you.
 * @return The default dbm filename.
 */
static char *make_default_db_filename(apr_pool_t *p)
{
    apr_pool_t *ptemp;
    char *tempdir;
    char *retval;

    apr_pool_create(&ptemp, p);

    tempdir = ap_server_root_relative(ptemp, "logs");
    apr_filepath_merge(&retval, tempdir, ATO_DBM_FILENAME, 0, p);

    apr_pool_destroy(ptemp);
  
    return retval;
} // char *make_default_db_filename



/*--------------------------------------------------------*
 * Module-level globals
 *--------------------------------------------------------*/

#define ERRSTR_SIZE 1000

/**
 * When someone first attempts to log in, and they're not in the database, we
 * need to say that we don't know if they're a good user yet.  Then, when we've
 * given the actual auth module a chance to authenticate them, we can check for
 * this value.
 * Note that using -88 here causes this value to become ambiguous (maybe it's a
 * legitimate timestamp) 88 seconds before the Epoch.  I don't know why you'd
 * set the server's system clock to that, but if you do, note this limitation.
 */
#define ATO_POSTPONE_CHECK -88

/**
 * For error reporting, we define an errstr.  
 * A place to store human-readable errors.
 */
static apr_status_t err;
static char errstr[ERRSTR_SIZE];

/**
 * Serial number -- How many times have we run?
 * We use that so that we can run sessions_cleanup every so-many runs.
 * This is shared-memory, so that the processes can have a global notion of how
 * many times it's been run.  Therefore, we protect it with a proc_mutex.
 */
static char *serial_lockfile;
static apr_proc_mutex_t *serial_lock;
static apr_shm_t *run_serial_shm;

/**
 * File that we will lock to lock access to the db.
 */
static apr_file_t *db_lockfile;


/*-------------------------------------------------------------*
 * Handlers for the configuration directives.
 *-------------------------------------------------------------*/

/**
 * Handle the AuthTimeOut configuration directive.
 * This sets the configuration of how many seconds of inactivity it takes to be
 * said to have timed-out of your connection.
 * @param cmd I don't know.  Required by the api.  I don't use it.
 * @param mconfig The effective config so far, cast to (void *).
 * @param word1 String rep. of how many secs of inactivity == timeout.
 * @param return NULL, or a string describing the problem.
 */
static const char *cmd_authtimeout(cmd_parms *cmd, void *mconfig, char *word1)
{
  x_cfg *cfg;
  char *ended_at;

  cfg = (x_cfg *)mconfig;
  cfg->timeout_time = strtol(word1, &ended_at, 10);

  if (ended_at != (word1 + strlen(word1))) {
      return "Bogus number given for AuthTimeout";
  } // if we didn't use the whole word in the making of the #

  return NULL;
} // static const char *cmd_authtimeout

/**
 * Handle the AuthTimeOutAnonTemplate configuration directive.
 * This tells us what file is the HTML template to send to anonymous users.  
 * The file should explain that, because they have accessed the page
 * anonymously, they will be prompted for their username and password the next
 * time they access the page.
 * It should also have a link that brings them back to the page they're on (a
 * cookie will keep track of whether or not they have seen the message).
 * <a href=""> should be a sufficient link.
 * @param cmd I don't know.  Required by the api.  I don't use it.
 * @param mconfig The effective config so far, cast to (void *).
 * @param word1 Filename of the template file.
 * @param return NULL, or a string describing the problem.
 */
static const char *cmd_anon_url(cmd_parms *cmd, void *mconfig, char *word1)
{
  x_cfg *cfg;

  cfg = (x_cfg *)mconfig;
  cfg->anon_url = apr_pstrdup(cfg->pool, word1);

  return NULL;
} // static const char *cmd_anon_url

/**
 * Handle the AuthTimeOutHelpfulUI configuration directive.
 * The helpful UI prints out the anon_url, anon_url type of things.
 * @param cmd I don't know.  Required by the api.  I don't use it.
 * @param mconfig The effective config so far, cast to (void *).
 * @param flag On or off.
 * @param return NULL, or a string describing the problem.
 */
static const char *cmd_authtimeouthelpfului(cmd_parms *cmd, void *mconfig, int flag)
{
  x_cfg *cfg;

  cfg = (x_cfg *)mconfig;
  cfg->helpful_ui = flag;

  return NULL;
} // static const char *cmd_authtimeouthelpfului


/**
 * Handle the AuthTimeOutNonAnonTemplate configuration directive.
 * This tells us what file is the HTML template to send to non-anonymous users
 * who have timed out.
 * The file should explain that, because their session has timed out, they will
 * be prompted for their username and password the next time they access the
 * page.
 * It should also have a link that brings them back to the page they're on (a
 * cookie will keep track of whether or not they have seen the message).
 * <a href=""> should be a sufficient link.
 * @param cmd I don't know.  Required by the api.  I don't use it.
 * @param mconfig The effective config so far, cast to (void *).
 * @param word1 Filename of the template file.
 * @param return NULL, or a string describing the problem.
 */
static const char *cmd_nonanon_url(
    cmd_parms *cmd, void *mconfig, char *word1)
{
  x_cfg *cfg;

  cfg = (x_cfg *)mconfig;
  cfg->nonanon_url = apr_pstrdup(cfg->pool, word1);

  return NULL;
} // static const char *cmd_anon_template


/**
 * Handle the AuthTimeOutCleanPeriod configuration directive.
 * This sets the conf. of how many runs it takes before we clean the database
 * of failed logins.
 * @param cmd I don't know.  Required by the api.  I don't use it.
 * @param mconfig The effective config so far, cast to (void *).
 * @param word1 String rep. of # -- how many runs till we clean the database.
 * @param return NULL, or a string describing the problem.
 */
static const char *cmd_authtimeoutcleanperiod(cmd_parms *cmd, void *mconfig, char *word1)
{
  x_cfg *cfg;
  char *ended_at;

  cfg = (x_cfg *)mconfig;
  cfg->clean_howoften = strtol(word1, &ended_at, 10);

  if (ended_at != (word1 + strlen(word1))) {
      return "Bogus number given for AuthTimeoutCleanPeriod";
  } // if we didn't use the whole word in the making of the #

  return NULL;
} // static const char *cmd_authtimeoutcleanperiod


/**
 * Handle the AuthTimeOutDBMFile configuration directive.
 * You can have multiple dbm files, but the key in each is:
 *   (client-ip, server-hostname, user, pass)
 * This directive sets what dbm file we're using to store the last-access times.
 * @param cmd I don't know.  Required by the api.  I don't use it.
 * @param mconfig The effective config so far, cast to (void *).
 * @param word1 The file-path of where we want the dbm file to be.
 * @param return NULL, or a string describing the problem.
 */
static const char *cmd_authtimeoutdbmfile(
                                cmd_parms *cmd, void *mconfig, char *word1)
{
  x_cfg *cfg;

  cfg = (x_cfg *)mconfig;
  cfg->file_for_db = apr_pstrdup(cfg->pool, word1);

  return NULL;
} // static const char *cmd_authtimeoutdbmfile

/**
 * Tell us what magic url will log you out.
 * @param cmd I don't know.  Required by the api.  I don't use it.
 * @param mconfig The effective config so far, cast to (void *).
 * @param word1 The magic url that will log you out.
 * @param return NULL, or a string describing the problem.
 */
static const char *cmd_authtimeoutlogouturl(cmd_parms *cmd, void *mconfig, char *word1) {
  x_cfg *cfg;

  cfg = (x_cfg *)mconfig;
  cfg->logout_url = apr_pstrdup(cfg->pool, word1);
  return NULL;
} // static const char *cmd_authtimeoutlogouturl

/**
 * Tell us the url to redirect you to if there is no Referrer set.
 * @param cmd I don't know.  Required by the api.  I don't use it.
 * @param mconfig The effective config so far, cast to (void *).
 * @param word1 The url to redirect you to.
 * @param return NULL, or a string describing the problem.
 */
static const char *cmd_authtimeoutredirectdefault(cmd_parms *cmd, void *mconfig, char *word1) {
  x_cfg *cfg;

  cfg = (x_cfg *)mconfig;
  cfg->logout_redirect_default = apr_pstrdup(cfg->pool, word1);
  return NULL;
} // static const char *cmd_authtimeoutredirectdefault

/**
 * Tell us the url to redirect you to even if there is a Referrer set.
 * @param cmd I don't know.  Required by the api.  I don't use it.
 * @param mconfig The effective config so far, cast to (void *).
 * @param word1 The url to redirect you to.
 * @param return NULL, or a string describing the problem.
 */
static const char *cmd_authtimeoutredirectforce(cmd_parms *cmd, void *mconfig, char *word1) {
  x_cfg *cfg;

  cfg = (x_cfg *)mconfig;
  cfg->logout_redirect_force = apr_pstrdup(cfg->pool, word1);
  return NULL;
} // static const char *cmd_authtimeoutredirectforce

/**
 * Tell us the file to display if they attempt to logout while not logged in.
 * @param cmd I don't know.  Required by the api.  I don't use it.
 * @param mconfig The effective config so far, cast to (void *).
 * @param word1 The url to redirect you to.
 * @param return NULL, or a string describing the problem.
 */
static const char *cmd_authtimeoutalreadyloggedouturl(cmd_parms *cmd, void *mconfig, char *word1) {
  x_cfg *cfg;

  cfg = (x_cfg *)mconfig;
  cfg->alreadyloggedout_url = apr_pstrdup(cfg->pool, word1);
  return NULL;
} // static const char *cmd_authtimeoutalreadyloggedouturl


/**
 * The commands table, so apache can run our configuration directives.
 */
static const command_rec x_cmds[] =
{
    AP_INIT_TAKE1(
        "AuthTimeOut",                      /* directive name */
        cmd_authtimeout,                    /* config action routine */
        NULL,                               /* argument to include in call */
        OR_OPTIONS,                         /* where available */
        "How long till the auth times out?" /* directive description */
    ),

    AP_INIT_TAKE1(
        "AuthTimeOutAnonMessage",
        cmd_anon_url,
        NULL,
        OR_OPTIONS,
        "A url that anonymous users are redirected to, warning that they will be prompted for a username and password."
    ),

    AP_INIT_TAKE1(
        "AuthTimeOutNonAnonMessage",
        cmd_nonanon_url,
        NULL,
        OR_OPTIONS,
        "A url that non-anonymous users who have timed out are redirected to, warning that they will be prompted again for a username and password."
    ),

    AP_INIT_TAKE1(
        "AuthTimeOutDBMFile",
        cmd_authtimeoutdbmfile,
        NULL,
        OR_OPTIONS,
        "Where to put the dbm file."
    ),

    AP_INIT_TAKE1(
        "AuthTimeOutCleanPeriod",
        cmd_authtimeoutcleanperiod,
        NULL,
        OR_OPTIONS,
        "Clean the auth-timeout database after this many runs."
    ),

    AP_INIT_TAKE1(
        "AuthTimeOutLogoutUrl",
        cmd_authtimeoutlogouturl,
        NULL,
        OR_OPTIONS,
        "Magic URL that immediately expires your auth session."
    ),

    AP_INIT_TAKE1(
        "AuthTimeOutLogoutRedirectDefault",
        cmd_authtimeoutredirectdefault,
        NULL,
        OR_OPTIONS,
        "URL to redirect you to after logout, if no Referrer is set."
    ),

    AP_INIT_TAKE1(
        "AuthTimeOutLogoutRedirectForce",
        cmd_authtimeoutredirectforce,
        NULL,
        OR_OPTIONS,
        "URL to redirect you to after logout, even if a Referrer is set."
    ),

    AP_INIT_TAKE1(
        "AuthTimeOutAlreadyLoggedOutUrl",
        cmd_authtimeoutalreadyloggedouturl,
        NULL,
        OR_OPTIONS,
        "URL to redirect you to if you attempt to log out while not logged in."
    ),

    AP_INIT_FLAG(
        "AuthTimeOutHelpfulUI",
        cmd_authtimeouthelpfului,
        NULL,
        OR_OPTIONS,
        "Explain what's going on during the login process?"
    ),

    {NULL}
};


/*
 * Declare ourselves so the configuration routines can find and know us.
 * We'll fill it in at the end of the module.
 */
module AP_MODULE_DECLARE_DATA auth_timeout_module;


/*--------------------------------------------------------------------------*/
/*                                                                          */
/* These routines are strictly internal to this module, and support its     */
/* operation.  They are not referenced by any external portion of the       */
/* server.                                                                  */
/*                                                                          */
/*--------------------------------------------------------------------------*/

/*
 * Locate our directory configuration record for the current request.
 * A cheap-and-dirty convenience function.
 */
static x_cfg *our_dconfig(const request_rec *r)
{
    return (x_cfg *) ap_get_module_config(r->per_dir_config, &auth_timeout_module);
}


/**
 * Authentication has failed.  Tell them so.
 * If we have determined that the auth has failed, this routine sends the
 * appropriate headers to make the browser decide to put the login box up
 * again.
 * @param r The request structure.
 * @return HTTP_UNAUTHORIZED is the only return value.
 */
static int auth_failed(request_rec *r)
{
  apr_table_setn(r->err_headers_out,
                 (PROXYREQ_PROXY == r->proxyreq) ? "Proxy-Authenticate"
                                                 : "WWW-Authenticate",
                 apr_pstrcat(r->pool, "Basic realm=\"", ap_auth_name(r),
                             "\"", NULL));
  return HTTP_UNAUTHORIZED;
}


/**
 * Get the basic auth information -- what user/pw did they use?
 * @param r The request structure.
 * @param[out] user The username they logged in with.
 * @param[out] pw The password they used.
 * @return HTTP_UNAUTHORIZED if they didn't attempt to log in.  OK else.
 */
static int get_basic_auth(request_rec *r, char **user, char **pw)
{
    const char *auth_line;
    char *decoded_line;
    int length;

    /* Get the appropriate header */
    auth_line = apr_table_get(r->headers_in, (PROXYREQ_PROXY == r->proxyreq)
                                              ? "Proxy-Authorization"
                                              : "Authorization");
    apr_pool_t *ptemp;
    char *tempdir;
    char *retval;

    if (!auth_line) {
        return auth_failed(r);
    }

    if (strcasecmp(ap_getword(r->pool, &auth_line, ' '), "Basic")) {
        /* Client tried to authenticate using wrong auth scheme */
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                      "client used wrong authentication scheme: %s", r->uri);
        return auth_failed(r);
    }

    /* Skip leading spaces. */
    while (isspace(*auth_line)) {
        auth_line++;
    }

    decoded_line = apr_palloc(r->pool, apr_base64_decode_len(auth_line) + 1);
    length = apr_base64_decode(decoded_line, auth_line);
    /* Null-terminate the string. */
    decoded_line[length] = '\0';

    *user = ap_getword_nulls(r->pool, (const char**)&decoded_line, ':');
    *pw = decoded_line;

    /* set the user, even though the user is unauthenticated at this point */
    // r->user = (char *) *user;

    return OK;
}


/**
 * Make keys to use to query the database.  Both the main key and the ts:: key.
 * The main key is used to see when the last time they successfully logged in
 * was.  If the value in the db at the main key is ATO_POSTPONE_CHECK, that
 * means that they attempted to log in, but since they weren't in the database,
 * we have no idea whether this will succeed or fail.  If we need to know when
 * they made that attempt, the answer is at ts::<key>.  
 * This routine creates both of those keys and stores them in datums.  They are
 * allocated out of the request pool.
 * @param r The request record.
 * @param[out] key_datum The key to see when their last successful attempt was.
 * @param[out] timestampkey_datum The key to see when they made an attempt.
 * @return 0 or http error.
 */
static int make_keys(
    request_rec *r, apr_datum_t *key_datum, apr_datum_t *timestampkey_datum)
{
    int res;
    int keylen;
    char *key;
    char *md5edkey;
    int md5edkeylen = 128;
    char *timestamp_key;
    char *sent_user;
    char *sent_pw;
    char *sname = r->server->server_hostname;

    sname = (sname != NULL) ? sname : "";

    res = get_basic_auth(r, &sent_user, &sent_pw);
    if (res) {
      return res;
    }

    // Construct the key
    keylen = strlen(sname) + strlen("::");
    keylen += strlen(r->connection->remote_ip) + strlen("::");
    keylen += strlen(sent_user) + strlen("::");
    keylen += strlen(sent_pw);
    keylen += 1; // keep the '\0'

    key = (char *)apr_palloc(r->pool, sizeof(char) * (keylen+1));
    sprintf(key, "%s::%s::%s::%s\0", 
        sname, r->connection->remote_ip, sent_user, sent_pw);

    md5edkey = (char *)apr_palloc(r->pool, sizeof(char) * md5edkeylen);

    apr_md5_encode(key, sent_user, md5edkey, md5edkeylen);
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, 
        "mod_auth_timeout: MD5ed(%s)", md5edkey);

    key_datum->dptr = md5edkey;
    key_datum->dsize = strlen(md5edkey)+1;

    keylen = key_datum->dsize + 4;
    timestamp_key = (char *)apr_palloc(r->pool, sizeof(char) * (keylen+1));
    sprintf(timestamp_key, "ts::%s\0", md5edkey);

    timestampkey_datum->dptr = timestamp_key;
    timestampkey_datum->dsize = keylen;

    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, 
        "mod_auth_timeout: KEY(%s)", timestamp_key);

    return 0;
} // int make_keys


/**
 * Open the DBM file for use in this request.
 * @param[out] dbm The dbm handle, once open.
 * @param our_config The config structure (says where to look for the file).
 * @param r The request structure.
 * @return OK or error.
 */
static apr_status_t open_db(apr_dbm_t **dbm, x_cfg *our_config, request_rec *r)
{
    err=apr_dbm_open(
              dbm, our_config->file_for_db, 
              APR_DBM_RWCREATE, APR_OS_DEFAULT, r->pool);
    if (err) {
      apr_strerror(err, errstr, ERRSTR_SIZE);
      ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL, 
          "mod_auth_timeout: Failed to open dbm file (%s).  Error: %s.", 
          our_config->file_for_db, 
          errstr);
      return err;
    } // if the open of the dbm file failed

    return OK;
} // apr_status_t open_db


/**
 * Lock the database.
 * There are several non-atomic database operations that we do.  Therefore,
 * they must be locked.  We lock only the file that we're using for this
 * request.
 * @param our_config The config structure.
 * @param r The request structure.
 * @return OK or error.
 */
static apr_status_t lock_db(x_cfg *our_config, request_rec *r)
{
  char *lockfile_name;

  lockfile_name = apr_pstrcat(r->pool, our_config->file_for_db, ".lock", NULL);
  err = apr_file_open(
      &db_lockfile, lockfile_name, 
      APR_CREATE|APR_WRITE|APR_BINARY|APR_DELONCLOSE, 0, r->pool);
  apr_file_perms_set(lockfile_name, APR_UREAD | APR_UWRITE | APR_GREAD | APR_GWRITE);

  if (err) {
    apr_strerror(err, errstr, ERRSTR_SIZE);
    ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL, 
        "mod_auth_timeout: Failed to create lock file (%s).  Error: %s.", 
        lockfile_name, 
        errstr);
    return err;
  } // if we couldn't open the file

  apr_file_lock(db_lockfile, APR_FLOCK_EXCLUSIVE);

  return OK;
} // apr_status_t lock_db


/**
 * Unlock the databases.
 * We only lock/unlock the actual db used for this request.
 * @param our_config The config struct that says which dbm file to use.
 * @param r The request record.
 * @return OK.
 */
static apr_status_t unlock_db(x_cfg *our_config, request_rec *r)
{
  apr_file_unlock(db_lockfile);
  apr_file_close(db_lockfile);

  return OK;
} // apr_status_t unlock_db


static apr_status_t close_db(apr_dbm_t *dbm, request_rec *r)
{
  apr_dbm_close(dbm);

  return OK;
} // close_db


/**
 * Construct a cookie to send, which will unset my previous cookies.
 * @return The text of a cookie.
 */
static char *killcookie(request_rec *r) {
  char *txt;
  apr_time_t the_before_times;

  txt = apr_pstrdup(r->pool, "Fri, 13 Jul XXXX XX:XX:XX UTC");
  the_before_times = apr_rfc822_date(
      txt, 
      apr_time_from_sec(apr_time_sec(apr_time_now()) - 86400));
  txt = apr_pstrcat(
      r->pool,
      "mod_auth_timeout=goodtogo ; expires=",
      txt,
      NULL);

  return txt;
} // char *killcookie

/**
 * Debugging routine that prints the contents of our database.
 * For some reason, I can't get it to work by printing to the debugging-level
 * log.  I print to the error log instead.  Because of that, I intend to leave
 * the call to this function commented out when not in development.  If you can
 * figure out how to make it go to the debugging log, it would me much
 * appreciated.
 * @param r The reuqest structure.
 */
static void print_what_we_have(request_rec *r)
{
  int howmany;
  int stopnow;

  apr_datum_t key_datum;
  apr_datum_t val_datum;

  int last_access_time;

  x_cfg *our_config = our_dconfig(r);

  apr_dbm_t *sessions;

  open_db(&sessions, our_config, r);
    
  // Enter critical section
  if ((err = lock_db(our_config, r))) {
    apr_strerror(err, errstr, ERRSTR_SIZE);
    ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL, 
        "mod_auth_timeout: Failed to open lock file.  Error: %s.", errstr);
  } // if the locking failed

  howmany=0;
  for (stopnow=apr_dbm_firstkey(sessions, &key_datum);
        !stopnow && key_datum.dsize;
        stopnow=apr_dbm_nextkey(sessions, &key_datum))
  {
    howmany += 1;
  } // for all the keys

  ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, "There are %d entries", howmany);

  for (stopnow=apr_dbm_firstkey(sessions, &key_datum);
        !stopnow && key_datum.dsize;
        stopnow=apr_dbm_nextkey(sessions, &key_datum))
  {
      apr_dbm_fetch(sessions, key_datum, &val_datum);
      memcpy(&last_access_time, val_datum.dptr, sizeof(int));

      ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, 
          "%s: %d", key_datum.dptr, last_access_time);
  } // for all the things in the hash table

  // End critical section
  unlock_db(our_config, r);
  apr_dbm_close(sessions);
} // void print_what_we_have


/**
 * Read the headers and parse each individual cookie into the cfg.
 */
static int cookiefy(void *arg, const char *key, const char *val) {
  x_cfg *cfg = (x_cfg *)arg;
  apr_pool_t *temp_pool;
  char *cookies_str;
  char *cookie_str;
  char *k;
  char *v;
  char *last = NULL;
  char *last2 = NULL;

  apr_pool_create(&temp_pool, NULL);
  cookies_str = apr_pstrdup(temp_pool, val);
  
  while ( (cookie_str=apr_strtok(cookies_str, ";", &last)) ) {
    cookies_str = NULL;
    k = apr_strtok(cookie_str, "=", &last2);
    v = apr_strtok(NULL, "=", &last2);

    apr_table_add(cfg->cookies, k, v);
  } // while there's cookies in the string

  apr_pool_destroy(temp_pool);
  
  return 1;
} // static int cookiefy

/**
 * Figure out what state the user is in.
 * They could be in ATO_ANON_IGNORANT
 * Figure out what state we're in.
 * There are XXX states:
 * 1. Anonymous & ignorant:
 *    They came in anonymously and have not been shown an information page.
 * 2. Anonymous & savvy:
 *    After coming in anonymously, they saw a page warning that they'd have
 *    to type in a password.  They came back anonymously.
 * 3. Non-anonymous & logged in.
 *    They came in with their username and password and were not timed out.
 *    Show the page.
 * 4. Non-anonymous & timed out & ignorant.
 *    They came in with a username/password, but were timed out.  Show them
 *    a page warning that they'll have to put in their username and
 *    password again.
 * 5. Non-anonymous & timed out & savvy.
 *    They came in with a username/password, but were timed out.  They saw
 *    the message saying they'd be prompted.  They're coming back.  We
 *    should give them their page.
 *
 * Here's what we do:
 * If they come in anonymously with no interesting cookies, then they are
 * in state 1.  
 *    We will send them a cookie mod_auth_timeout=anon and display the
 *    appropriate page.
 *
 * If they come in anonymously with mod_auth_timeout=anon, then they are in
 * state 2.  
 *    We will send them UNAUTHORIZED and expire the mod_auth_timeout cookie.
 *
 * If they come in with a username and password, check if they're timed
 * out.  If they are not, they're in state 3.
 *    We will send them the page they want.
 *
 * If they come in with a username and password, check if they're timed
 * out.  If they are, they're in state 4.
 *    Send them a cookie mod_auth_timeout=timedout and display a page
 *    warning them that they're about to be prompted.
 *
 * If they come in with a username and password, and
 * mod_auth_timeout=timedout, then check if they're timed out.  If they
 * are, they are in state 4.  If not, they are in state 5.
 *    We should expire mod_auth_timeout and give them their page.
 *
 * Note that, in this function, states 3, 4, and 5 cannot be fully
 * distinguished from one another.  We need more information, that is the
 * output of two hooks working together.  So we can only return these 4 things:
 *   ATO_STATE_ANON_IGNORANT
 *   ATO_STATE_ANON_SAVVY
 *   ATO_STATE_NONANON_IGNORANT
 *   ATO_STATE_NONANON_SAVVY
 * You'll have to use further clues from your context in order to further
 * distinguish.
 */
#define ATO_STATE_ANON_IGNORANT       1
#define ATO_STATE_ANON_SAVVY          2
#define ATO_STATE_NONANON_IGNORANT    3
#define ATO_STATE_NONANON_SAVVY       4
static int determine_state(request_rec *r) {
  x_cfg *cfg = our_dconfig(r);
  char *txtval;
  char *sent_user;
  char *sent_pw;
  int anon;
  int res;

  if (cfg->cookies == NULL) {
    cfg->cookies = apr_table_make(r->pool, 1);
    apr_table_do(cookiefy, cfg, r->headers_in, "Cookie", NULL);
  } // if the cookies table is not made, let's make it


  // Are they anonymous?
  res = get_basic_auth(r, &sent_user, &sent_pw);
  anon = (res != OK);

  // What's their cookie?
  txtval = apr_table_get(cfg->cookies, "mod_auth_timeout");

  if (anon) {
    if (txtval && !strcmp(txtval, "anon")) {
      return ATO_STATE_ANON_SAVVY;
    } // if they're savvy
    else {
      return ATO_STATE_ANON_IGNORANT;
    } // else they're ignorant
  } // if they're anonymous, are they savvy?

  // I assert that you are logged in with a username and password.

  if (txtval && !strcmp(txtval, "timedout")) {
    return ATO_STATE_NONANON_SAVVY;
  } // if they are savvy about being timed out
  else {
    return ATO_STATE_NONANON_IGNORANT;
  } // else they are not savvy about it.
} // int determine_state


/**
 * Sometimes run.  Free memory from expired sessions.
 * It deletes unsuccessful logins from the database after awhile.
 * It does not delete expired successful logins until server restart, because
 * that would be a security concern:  You could just leave your browser open
 * and then request again after you got deleted from the database.
 * However, this behavior is also a concern, because the database could get
 * filled up with a different row for each (ip, user) pair that's ever
 * successfully logged in.
 * @param r The request structure.
 */
static int sessions_cleanup(request_rec *r)
{
  // After how many runs do we run the cleanup?
  int *run_serial;
  int run_often;

  int stopnow;

  apr_datum_t key_datum;
  apr_datum_t key2_datum;
  apr_datum_t val_datum;

  apr_dbm_t *sessions;

  int last_access_time;

  x_cfg *our_config = our_dconfig(r);

  run_often = our_config->clean_howoften;

  ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, "mod_auth_timeout:  my timeout(%d)", our_config->timeout_time);

  if (open_db(&sessions, our_config, r)) {
    ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL, 
        "mod_auth_timeout: Failed to open dbm file (%s).  Error: %s.", 
        our_config->file_for_db, 
        errstr);
    return HTTP_INTERNAL_SERVER_ERROR;
  } // if failed opening db


  // print_what_we_have(r);

  apr_proc_mutex_lock(serial_lock);
  run_serial = (int *)apr_shm_baseaddr_get(run_serial_shm);
  run_serial[0] = (run_serial[0] + 1) % run_often;
  apr_proc_mutex_unlock(serial_lock);
  if (run_serial[0] != 0) {
    apr_dbm_close(sessions);
    return OK;
  } // if we will not run this time
  else {
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, "mod_auth_timeout:  Cleaning up the sessions table.");
  } // else we will run

  if ((err = lock_db(our_config, r))) {
    apr_strerror(err, errstr, ERRSTR_SIZE);
    ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL, 
        "mod_auth_timeout: Failed to open lock file.  Error: %s.", errstr);
    apr_dbm_close(sessions);
    return err;
  } // if the locking failed

  for (stopnow=apr_dbm_firstkey(sessions, &key_datum);
        !stopnow && key_datum.dsize;
        stopnow=apr_dbm_nextkey(sessions, &key_datum))
  {
      apr_dbm_fetch(sessions, key_datum, &val_datum);
      memcpy(&last_access_time, val_datum.dptr, sizeof(int));

      // We will loop through and delete the things we don't need anymore.
      // Which is to say, we will delete expired things.  In the case of
      // POSTPONEd things, we will delete them iff their associated ts:: key
      // (which is a timestamp) is expired.
      if (last_access_time == ATO_POSTPONE_CHECK) {
          // Don't delete it unless the ts:: key tells us to.
          continue;
      } // if it's a postponed one

      if (key_datum.dsize > 4) {
        if (!strncmp(key_datum.dptr, "ts::", 4)) {
          if (apr_time_sec(r->request_time) - last_access_time 
                                                  > our_config->timeout_time) 
          {
            key2_datum.dsize = key_datum.dsize - 4;
            key2_datum.dptr = (char *)apr_palloc(
                  r->pool, sizeof(char) * key2_datum.dsize);
            memcpy(key2_datum.dptr, key_datum.dptr + 4, key2_datum.dsize);

            ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, "Deleting %s",
                key_datum.dptr);
            ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, "Deleting %s",
                key2_datum.dptr);

            apr_dbm_delete(sessions, key_datum);
            apr_dbm_delete(sessions, key2_datum);

            continue;
          } // if they failed to log in too long ago
        } // if it's a timestamp one
      } // if we can check this for ts::
  } // for all the things in the hash table
  unlock_db(our_config, r);

  apr_dbm_close(sessions);
  return OK;
} // void sessions_cleanup


/*--------------------------------------------------------------------------*/
/*                                                                          */
/* Now let's declare routines for each of the callback phase in order.      */
/* (That's the order in which they're listed in the callback list, *not     */
/* the order in which the server calls them!  See the command_rec           */
/* declaration near the bottom of this file.)  Note that these may be       */
/* called for situations that don't relate primarily to our function - in   */
/* other words, the fixup handler shouldn't assume that the request has     */
/* to do with "example" stuff.                                              */
/*                                                                          */
/* With the exception of the content handler, all of our routines will be   */
/* called for each request, unless an earlier handler from another module   */
/* aborted the sequence.                                                    */
/*                                                                          */
/* Handlers that are declared as "int" can return the following:            */
/*                                                                          */
/*  OK          Handler accepted the request and did its thing with it.     */
/*  DECLINED    Handler took no action.                                     */
/*  HTTP_mumble Handler looked at request and found it wanting.             */
/*                                                                          */
/* What the server does after calling a module handler depends upon the     */
/* handler's return value.  In all cases, if the handler returns            */
/* DECLINED, the server will continue to the next module with an handler    */
/* for the current phase.  However, if the handler return a non-OK,         */
/* non-DECLINED status, the server aborts the request right there.  If      */
/* the handler returns OK, the server's next action is phase-specific;      */
/* see the individual handler comments below for details.                   */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*
 * This function gets called to create a per-directory configuration
 * record.  This will be called for the "default" server environment, and for
 * each directory for which the parser finds any of our directives applicable.
 * If a directory doesn't have any of our directives involved (i.e., they
 * aren't in the .htaccess file, or a <Location>, <Directory>, or related
 * block), this routine will *not* be called - the configuration for the
 * closest ancestor is used.
 *
 * The return value is a pointer to the created module-specific
 * structure.
 */
static void *x_create_dir_config(apr_pool_t *p, char *dirspec)
{
    x_cfg *cfg;

    /*
     * Allocate the space for our record from the pool supplied.
     */
    cfg = (x_cfg *) apr_pcalloc(p, sizeof(x_cfg));
    cfg->pool = p;
    /*
     * Now fill in the defaults.  If there are any `parent' configuration
     * records, they'll get merged as part of a separate callback.
     */
    cfg->cmode = CONFIG_MODE_DIRECTORY;
    cfg->timeout_time = ATO_TIMEOUT;
    cfg->file_for_db = make_default_db_filename(p);
    cfg->clean_howoften = ATO_CLEAN_HOWOFTEN;
    cfg->logout_url = NULL;
    cfg->logout_redirect_default = NULL;
    cfg->logout_redirect_force = NULL;
    cfg->alreadyloggedout_url = NULL;
	cfg->helpful_ui = ATO_HELPFULUI;

    return (void *) cfg;
}

/*
 * This function gets called to create a per-server configuration
 * record.  It will always be called for the "default" server.
 *
 * The return value is a pointer to the created module-specific
 * structure.
 */
static void *x_create_server_config(apr_pool_t *p, server_rec *s)
{
    x_cfg *cfg;

    /*
     * As with the x_create_dir_config() reoutine, we allocate and fill
     * in an empty record.
     */
    cfg = (x_cfg *) apr_pcalloc(p, sizeof(x_cfg));
    cfg->pool = p;
    cfg->cmode = CONFIG_MODE_SERVER;
    cfg->timeout_time = ATO_TIMEOUT;
    cfg->file_for_db = make_default_db_filename(p);
    cfg->clean_howoften = ATO_CLEAN_HOWOFTEN;
    cfg->logout_url = NULL;
    cfg->logout_redirect_default = NULL;
    cfg->logout_redirect_force = NULL;
    cfg->alreadyloggedout_url = NULL;
	cfg->helpful_ui = ATO_HELPFULUI;

    return (void *) cfg;
}

/*
 * This function gets called to merge two per-server configuration
 * records.  This is typically done to cope with things like virtual hosts and
 * the default server configuration  The routine has the responsibility of
 * creating a new record and merging the contents of the other two into it
 * appropriately.  If the module doesn't declare a merge routine, the more
 * specific existing record is used exclusively.
 *
 * The routine MUST NOT modify any of its arguments!
 *
 * The return value is a pointer to the created module-specific structure
 * containing the merged values.
 */
static void *x_merge_config(apr_pool_t *p, void *server1_conf,
                                         void *server2_conf)
{
    x_cfg *merged_config = (x_cfg *) apr_palloc(p, sizeof(x_cfg));
    x_cfg *s1conf = (x_cfg *) server1_conf;
    x_cfg *s2conf = (x_cfg *) server2_conf;

    memcpy(merged_config, s1conf, sizeof(x_cfg));
    /*
     * Our inheritance rules are our own, and part of our module's semantics.
     * Basically, just note whence we came.
     */
    merged_config->cmode =
        (s1conf->cmode == s2conf->cmode) ? s1conf->cmode : CONFIG_MODE_COMBO;
    merged_config->loc = apr_pstrdup(p, s2conf->loc);

    if (s2conf->timeout_time) {
      merged_config->timeout_time = s2conf->timeout_time;
    } // if the more specific one has a new idea what timeout time they want

    if (s2conf->file_for_db) {
      merged_config->file_for_db = apr_pstrdup(p, s2conf->file_for_db);
    } // if the more specific one has a new idea of what file they wanna use

    if (s2conf->clean_howoften) {
      merged_config->clean_howoften = s2conf->clean_howoften;
    } // if more specific one sez

    if (s2conf->db_persists) {
      merged_config->db_persists = s2conf->db_persists;
    } // if more specific sez

    if (s2conf->logout_url) {
      merged_config->logout_url = apr_pstrdup(p, s2conf->logout_url);
    } // if more specific sez

    if (s2conf->logout_redirect_default) {
      merged_config->logout_redirect_default = s2conf->logout_redirect_default;
    } // if more specific sez

    if (s2conf->logout_redirect_force) {
      merged_config->logout_redirect_force = s2conf->logout_redirect_force;
    } // if more specific sez

    if (s2conf->alreadyloggedout_url) {
      merged_config->alreadyloggedout_url = s2conf->alreadyloggedout_url;
    } // if more specific sez

    if (s2conf->anon_url) {
      merged_config->anon_url = s2conf->anon_url;
    } // if more specific one sez

    if (s2conf->anon_url) {
      merged_config->nonanon_url = s2conf->nonanon_url;
    } // if more specific one sez

    merged_config->helpful_ui = s2conf->helpful_ui;

    return (void *) merged_config;
}

/*
 * This routine is called to perform any module-specific fixing of header
 * fields, et cetera.  It is invoked just before any content-handler.
 *
 * The return value is OK, DECLINED, or HTTP_mumble.  If we return OK, the
 * server will still call any remaining modules with an handler for this
 * phase.
 */
static int x_post_config(apr_pool_t *pconf, apr_pool_t *plog,
                          apr_pool_t *ptemp, server_rec *s)
{
    char *tmpstr;
    const char *tmpdir;
    int *run_serial;

    apr_temp_dir_get(&tmpdir, ptemp);

    // Create 2 things: one locks and one shared-mem.
    // 1. Create the lock for the serial number.
    apr_filepath_merge(&tmpstr, tmpdir, "ato_serlock", 0, ptemp);
    serial_lockfile = ap_append_pid(pconf, tmpstr, ".");
    apr_proc_mutex_create(
        &serial_lock, serial_lockfile, APR_LOCK_DEFAULT, pconf);

    // 2. Create the shared-memory for the serial number.
    apr_shm_create(&run_serial_shm, sizeof(int), NULL, pconf);
    //   (2a) initialize the shared-memory for the serial number.
    run_serial = (int *)apr_shm_baseaddr_get(run_serial_shm);
    run_serial[0] = 0;

    return OK;
}

/**
 * This has gotten each "Cookie:" header.  Add its contents to the string.
 */
static int print_cookies(void *arg, const char *key, const char *val) {
  apr_pool_t *p = ((void **)arg)[0];
  char **str = ((void **)arg)[1];

  if (*str) {
    *str = apr_pstrcat(p, *str, "Cookie(", val, ")<br />\n", NULL);
  } // if we had an old string to append to
  else {
    *str = apr_pstrcat(p, "Cookie(", val, ")<br />\n", NULL);
  } // else we are starting output

  return 1;
} // static int print_cookies


/**
 * Set up the redirect if they are logging out.  Return 0 if they are not.
 */
int process_logout_redirect(x_cfg *our_config, request_rec *r) {
	char *redirect_url;
	if (our_config->logout_url && !strcmp(r->uri, our_config->logout_url)) {
		if (our_config->logout_redirect_force) {
			redirect_url = our_config->logout_redirect_force;
		} else {
			redirect_url = apr_table_get(r->headers_in, "Referrer");
			if (!redirect_url) {
				redirect_url = our_config->logout_redirect_default;
			}
		}
		if (redirect_url) {
			apr_table_setn(r->headers_out, "Location", redirect_url);
			return HTTP_MOVED_TEMPORARILY;
		}
	}

	return 0;
}


/**
 * Do the correct thing if they are logging out when not logged in.
 * Return 0 if this is not applicable.
 */
int process_unnecessary_logout(x_cfg *our_config, request_rec *r) {
	char *redirect_url = NULL;
	if (our_config->logout_url && !strcmp(r->uri, our_config->logout_url)) {
		if (our_config->alreadyloggedout_url) {
			redirect_url = our_config->alreadyloggedout_url;
		}

		apr_table_set(r->notes, "logout_state", "unnecessary");
		if (redirect_url) {
			apr_table_setn(r->headers_out, "Location", redirect_url);
			return HTTP_MOVED_TEMPORARILY;
		} else {
			// Let our handler show them some stuff.
			return OK;
		}
	}

	return DECLINED;
}




/*
 * This routine is called to check the authentication information sent with
 * the request (such as looking up the user in a database and verifying that
 * the [encrypted] password sent matches the one in the database).
 *
 * The return value is OK, DECLINED, or some HTTP_mumble error (typically
 * HTTP_UNAUTHORIZED).  If we return OK, no other modules are given a chance
 * at the request during this phase.
 */
static int x_check_user_id(request_rec *r)
{
    const char *current_auth_type;
    int not_in_db;
    int last_access_time;
    int current_time;
    int timestamp;
    int require_auth = 0;
    int were_tentative = 0;
    apr_status_t errthis;

    char *cookieval;

    apr_datum_t key_datum;
    apr_datum_t val_datum;

    apr_datum_t timestampkey_datum;
    apr_datum_t timestampval_datum;

    apr_dbm_t *sessions = NULL;
    int state;

    x_cfg *our_config = our_dconfig(r);

    // If they don't want timeout, they don't get it.
    if (our_config->timeout_time == 0) {
        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, 
            "mod_auth_timeout: No timeout.");
        return DECLINED;
    }

    apr_table_set(r->notes, "check_user_id", "called");

    if (open_db(&sessions, our_config, r)) {
      ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, 
          "mod_auth_timeout:check_user_id: Failed to open dbm file.");
      return HTTP_INTERNAL_SERVER_ERROR;
    } // if open sessions failed


    if ((errthis = lock_db(our_config, r))) {
      apr_strerror(err, errstr, ERRSTR_SIZE);
      ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL, 
          "mod_auth_timeout: Failed to open lock file.  Error: %s.", errstr);

      close_db(sessions, r);
      return err;
    } // if the locking failed

    unlock_db(our_config, r);


    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, 
        "mod_auth_timeout: BEFORE:::");
    // print_what_we_have(sessions, r);

    /* Are we configured to use some auth? */
    current_auth_type = ap_auth_type(r);
    if (!current_auth_type || !strcasecmp(current_auth_type, "None")) {
        close_db(sessions, r);
        return DECLINED;
    }

    state = determine_state(r);

    if (make_keys(r, &key_datum, &timestampkey_datum)) {
		int retval;
        // The most likely reason to fail to make a key is that they didn't
        // supply a username or password in their request.
        close_db(sessions, r);


		retval = process_unnecessary_logout(our_config, r);
		if (retval != DECLINED) {
		  apr_table_set(r->notes, "logout_state", "unnecessary");
		  ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, 
			  "mod_auth_timeout: Unnecessary logout 0? (%d)", retval);
		  return retval;
		}

        switch (state) {
        case ATO_STATE_ANON_IGNORANT:
          // If they're ignorant, we need to show them the page that tells them
          // about how they're anonymous.
          if (our_config->helpful_ui) {
			  return OK;
		  } else {
		      return DECLINED;
		  }
          break;

        case ATO_STATE_ANON_SAVVY:
          // If they're savvy, we can take them right to their page (where, I
          // trust, they will be UNAUTHORIZED.
          return DECLINED;
          break;
        
        default:
          ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL, 
              "mod_auth_timeout: make_keys failed for mysterious reasons.");
          return DECLINED;
          break;
        } // switch state
    } // if we failed to make a key

    // Enter critical section
    if ((errthis = lock_db(our_config, r))) {
      apr_strerror(err, errstr, ERRSTR_SIZE);
      ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL, 
          "mod_auth_timeout: Failed to open lock file.  Error: %s.", errstr);

      close_db(sessions, r);
      return err;
    } // if the locking failed

    // Check their timeout status
    require_auth = 0;
    not_in_db = 0;

    if (!apr_dbm_exists(sessions, key_datum)) {
      require_auth = 0;
      not_in_db = 1;
    } else {
      apr_dbm_fetch(sessions, key_datum, &val_datum);
      memcpy(&last_access_time, val_datum.dptr, sizeof(int));

      if (last_access_time == ATO_POSTPONE_CHECK) {
        were_tentative = 1;

        if (!apr_dbm_exists(sessions, timestampkey_datum)) {
          // This is totally absurd.  It should be the case that, when I put a
          // -88 in there, I also put the timestampkey in there.  If not,
          // something is effed up.
          ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, 
              "mod_auth_timeout:%d Timestampkey (%s) impossibly found not to be in db.", getpid(), timestampkey_datum.dptr);
          unlock_db(our_config, r);
          close_db(sessions, r);
          return HTTP_INTERNAL_SERVER_ERROR;
        } // if the timestampkey doesn't exist
        apr_dbm_fetch(sessions, timestampkey_datum, &val_datum);
        memcpy(&last_access_time, val_datum.dptr, sizeof(int));
      } // if they were marked as tentative
      
      current_time = apr_time_sec(r->request_time);
      if ((current_time - last_access_time) > our_config->timeout_time
          || state == ATO_STATE_NONANON_SAVVY) 
      {
        cookieval = killcookie(r);
        apr_table_add(r->err_headers_out, "Set-Cookie", cookieval);
        require_auth = 1;
      } // if they timed out
    } // if whether they are mentioned in the db


    // Mark now as the last time they accessed a page.
    // If they weren't in the database, we'll pass it on to the auth checker
    // and later (in ap_hook_check_access) check whether they succeeded.  If
    // they succeeded, we'll insert them into the database.
    if (not_in_db) {
      last_access_time = ATO_POSTPONE_CHECK;
    } else {
      last_access_time = apr_time_sec(r->request_time);
    } // if whether they were in the db

    val_datum.dptr = (char *)apr_palloc(r->pool, sizeof(int));
    memcpy(val_datum.dptr, &last_access_time, sizeof(int));
    val_datum.dsize = sizeof(last_access_time);

    // If we know that they timed out, let's delete them from the database and
    // then require auth.  That way, they'll put in their credentials and be
    // checked like someone who's never been there before -- this will get
    // around the problem I was having before, where you would have to log in
    // to set the timestamp, and then log in again to actually log in.
    if (!require_auth) {
      ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, 
          "mod_auth_timeout: They were fresh or not timed out.");

      errthis = apr_dbm_store(sessions, key_datum, val_datum);
      if (errthis) {
        apr_strerror(err, errstr, ERRSTR_SIZE);
        ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL, 
            "mod_auth_timeout: Failed to store in the dbm file (%s).  Error: %s.", 
            our_config->file_for_db, 
            errstr);
      }

      // If they were marked as tentative, we can get rid of their ts:: entry
      // now.
      if (were_tentative) {
        apr_dbm_delete(sessions, timestampkey_datum);
        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, 
            "mod_auth_timeout: Deleting4 (%s[%d]).", 
            timestampkey_datum.dptr,
            timestampkey_datum.dsize);
      } // if they were tentative

      // Let's see:  If they are attempting to log out, then let's do it.
      if (our_config->logout_url && !strcmp(r->uri, our_config->logout_url)) {
		// If they were trying to log out, this is where we let them do it.
		// Set them to being timed out immediately.
		last_access_time = apr_time_sec(r->request_time) 
							- our_config->timeout_time;
		memcpy(val_datum.dptr, &last_access_time, sizeof(int));
        errthis = apr_dbm_store(sessions, key_datum, val_datum);
        if (errthis) {
          apr_strerror(err, errstr, ERRSTR_SIZE);
          ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL, 
            "mod_auth_timeout: Failed to store in the dbm file (%s).  Error: %s.", 
            our_config->file_for_db, 
            errstr);
        }

        apr_table_add(r->notes, "logout_state", "logged out");
      }
    } else {
      ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, 
          "mod_auth_timeout: They were in the db but had timed out.");

      if (state == ATO_STATE_NONANON_SAVVY || !our_config->helpful_ui) {
        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, 
            "mod_auth_timeout: Deleting2 (%s).", key_datum.dptr);
        apr_dbm_delete(sessions, key_datum);

        if (were_tentative) {
          apr_dbm_delete(sessions, timestampkey_datum);

          ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, 
              "mod_auth_timeout: Deleting5 (%s[%d]).", 
              timestampkey_datum.dptr,
              timestampkey_datum.dsize);
        } // if they were tentative
      } // if they've seen everything they need to see
    } // if whether they timed out

    if (not_in_db) {
      ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, 
          "mod_auth_timeout: They weren't in the db.  We marked tentative.");

      timestamp = apr_time_sec(r->request_time);
      timestampval_datum.dptr = (char *)apr_palloc(r->pool, sizeof(int));
      memcpy(timestampval_datum.dptr, &timestamp, sizeof(int));
      timestampval_datum.dsize = sizeof(timestamp);

      errthis = apr_dbm_store(sessions, timestampkey_datum, timestampval_datum);
      if (errthis) {
        apr_strerror(err, errstr, ERRSTR_SIZE);
        ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL, 
            "mod_auth_timeout: Failed to store in the dbm file (%s).  Error: %s.", 
            our_config->file_for_db, 
            errstr);
      }

    } // if they weren't in the db

    // End critical section
    unlock_db(our_config, r);

    if ( (errthis = sessions_cleanup(r)) ) {
      apr_dbm_close(sessions);
      return errthis;
    } // if it messed up

    if (require_auth) {
      int retval;
      ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, 
          "mod_auth_timeout: Insisting on auth.");

      // I think this means that we know for sure that they have timed out.
      // Let's show them the 'you timed out' page if they're ignorant.
      close_db(sessions, r);

      retval = process_unnecessary_logout(our_config, r);
      if (retval != DECLINED) {
		  apr_table_set(r->notes, "logout_state", "unnecessary");
		  ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, 
			  "mod_auth_timeout: Unnecessary logout 1? (%d)", retval);
        return retval;
      }

      if (!our_config->helpful_ui) {
        // No special treatment.  We're not doing helpful ui.
        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, 
            "mod_auth_timeout: Insisting on auth.");
        return auth_failed(r);
      }

      switch(state) {
      case ATO_STATE_NONANON_IGNORANT:
        // Send a message to the content-handler.
        apr_table_add(r->notes, "timed-out", "true");
        // Allow them to get to the content-handler.
        return DECLINED;

      case ATO_STATE_NONANON_SAVVY:
        // No special treatment.  They knew what they were getting into.
        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, 
            "mod_auth_timeout: Insisting on auth.");
        return auth_failed(r);
        break;

      default:
        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, 
            "mod_auth_timeout: Nonsense state achieved.");
        break;
      } // switch on the state
    } // if require auth

    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, 
        "mod_auth_timeout: Passing them on.");

    close_db(sessions, r);
    return DECLINED;
} // x_check_user_id


/*
 * This routine is called to perform any module-specific fixing of header
 * fields, et cetera.  It is invoked just before any content-handler.
 *
 * If this is running, you can be assured that access was granted.
 *
 * The return value is OK, DECLINED, or HTTP_mumble.  If we return OK, the
 * server will still call any remaining modules with an handler for this
 * phase.
 */
static int x_fixer_upper(request_rec *r)
{
    int turnsout_they_were_good;
    apr_datum_t key_datum;
    apr_datum_t val_datum;

    apr_datum_t timestampkey_datum;

    x_cfg *our_config = our_dconfig(r);

    apr_dbm_t *sessions;
    apr_status_t errthis;

    int last_access_time;
	int retval;

    // If they don't want timeout, they don't get it.
    if (our_config->timeout_time == 0) {
        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, 
            "mod_auth_timeout: No timeout.");
        return DECLINED;
    }

    if (make_keys(r, &key_datum, &timestampkey_datum)) {
      int retval;
      // The most likely reason to fail to make a key is that they didn't put a
      // username or password in their request.
      
      retval = process_unnecessary_logout(our_config, r);
      if (retval != DECLINED) {
		  apr_table_set(r->notes, "logout_state", "unnecessary");
		  ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, 
			  "mod_auth_timeout: Unnecessary logout 2? (%d)", retval);
        return retval;
      }

      return DECLINED;
    } // if we failed to make a key

    if (open_db(&sessions, our_config, r)) {
      ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, 
          "mod_auth_timeout:fixups: Failed to open dbm file.");
      return HTTP_INTERNAL_SERVER_ERROR;
    } // if we failed to open db

    turnsout_they_were_good = 0;
    if (apr_dbm_exists(sessions, key_datum)) {
      apr_dbm_fetch(sessions, key_datum, &val_datum);
      memcpy(&last_access_time, val_datum.dptr, sizeof(int));
      
      if (last_access_time == ATO_POSTPONE_CHECK) {
        turnsout_they_were_good = 1;
      } // if they are new to the db
    } // if they are in the db


    if (!turnsout_they_were_good) {
      unlock_db(our_config, r);
      close_db(sessions, r);

      return DECLINED;
    }

    // Turns out, they were good.
    if (apr_dbm_exists(sessions, timestampkey_datum)) {
      ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, 
          "mod_auth_timeout: Deleting3 (%s).", timestampkey_datum.dptr);
      apr_dbm_delete(sessions, timestampkey_datum);
    } // if they had a timestamp in there

	if (our_config->logout_url && !strcmp(r->uri, our_config->logout_url)) {
		// If they were trying to log out, this is where we let them do it.
		// Set them to being timed out immediately.
		last_access_time = apr_time_sec(r->request_time) 
							- our_config->timeout_time;
        apr_table_add(r->notes, "logout_state", "logged out");
	} else {
		// Mark that they successfully logged in now.
		last_access_time = apr_time_sec(r->request_time);
	}
    val_datum.dptr = (char *)apr_palloc(r->pool, sizeof(int));
    memcpy(val_datum.dptr, &last_access_time, sizeof(int));
    val_datum.dsize = sizeof(last_access_time);
    errthis = apr_dbm_store(sessions, key_datum, val_datum);
    if (errthis) {
      apr_strerror(err, errstr, ERRSTR_SIZE);
      ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL, 
          "mod_auth_timeout: Failed to store in the dbm file (%s).  Error: %s.", 
          our_config->file_for_db, 
          errstr);
    }

    unlock_db(our_config, r);

    close_db(sessions, r);

    return OK;
}


/**
 * The content handler will respond to all URLs.
 * If you are anonymous, it will give you an info page.  Clicking a link there
 * will give you UNAUTHORIZED, so you can type your username and password.
 * If you are timed out, it will give you a different info page, with another
 * link to click on.  That will give you UNAUTHORIZED again so you can type
 * your username and password.
 * Finally, you are brought to the real page (hopefully).
 * State is advanced by cookies rather than successive GET or POST requests, so
 * as to not interfere with, or intercept, application logic.
 */
static int x_handler(request_rec *r)
{
    x_cfg *our_config;
    apr_file_t *template;
    apr_finfo_t info;
    int filesize;
    char *retval;
    char *default_txt;
    char *txt;
    int state;
    int showfile;
    int use_default_text;
    char *cookieval;
    int sendcookie;
    int decline;
	int unnecessary_logout;
	char *logout_state = NULL;

    our_config = our_dconfig(r);

	// We check if we're at the magic logout url BEFORE checking helpful-ui,
	// because responding to logout has nothing to do with helpful-ui -- It's
	// bare-minimum UI.
	// Also note that we only get here if we didn't get redirected on logout.
	logout_state = apr_table_get(r->notes, "logout_state");
	if (logout_state && !strcmp(logout_state, "unnecessary")) {
		// Send a header
		ap_set_content_type(r, "text/html");

		/*
		 * If we're only supposed to send header information (HEAD request),
		 * we're already there.
		 */
		if (r->header_only) {
			return OK;
		}

		default_txt = apr_pstrcat(
			r->pool,
			DOCTYPE_XHTML_1_0S,
			"<html>\n",
			"<head>\n",
			"<title>Already Logged Out</title>\n",
			"</head>\n",
			"<body>\n",
			"<h1>Already Logged Out</h1>\n\n",
			"<p>You attempted to log out.  ",
			"However, you were not logged in.</p>",
			"</body>\n",
			"</html>\n",
			NULL);
		ap_rputs(default_txt, r);
		return OK;
	} else if (logout_state && !strcmp(logout_state, "logged out")) {
		// Again, if we're getting here, it's because they didn't redirect you
		// anywhere.
		ap_set_content_type(r, "text/html");
		default_txt = apr_pstrcat(
			r->pool,
			DOCTYPE_XHTML_1_0S,
			"<html>\n",
			"<head>\n",
			"<title>Successfully Logged Out</title>\n",
			"</head>\n",
			"<body>\n",
			"<h1>Successfully Logged Out</h1>\n\n",
			"<p>You have been logged out.</p>",
			"</body>\n",
			"</html>\n",
			NULL);
		ap_rputs(default_txt, r);
		return OK;
	}

    if (!our_config->helpful_ui) {
      return DECLINED;
    } // if we should send them to the real content handler

    // We will leave them alone if we have no business on this request.
    txt = apr_table_get(r->notes, "check_user_id");
    if (!txt || strcmp(apr_table_get(r->notes, "check_user_id"), "called")) {
      return DECLINED;
    } // if we are not using any authentication scheme

    state = determine_state(r);

    // Let's see if they need a cookie!
    sendcookie = 0;
    decline = 0;

	switch (state) {
	case ATO_STATE_ANON_IGNORANT:
	  cookieval = "mod_auth_timeout=anon";
	  sendcookie = 1;
	  break;

	case ATO_STATE_NONANON_IGNORANT:
	case ATO_STATE_NONANON_SAVVY:
	  // They're not anonymous.  Are they logged in or timed out?
	  txt = apr_table_get(r->notes, "timed-out");
	  if (!txt || strcmp(txt, "true")) {
		// They are not timed out.  Give them the real page.
		// Also, expire their cookie.
		cookieval = killcookie(r);
		sendcookie = 1;
		decline = 1;
	  } // if they're not timed out, then they're logged in!
	  else {
		// They are timed out.  Give them the cookie.
		cookieval = "mod_auth_timeout=timedout";
		if (state == ATO_STATE_NONANON_IGNORANT) {
		  sendcookie = 1;
		} // if they were ignorant, make them not be
	  } // else they are timed out

	  break;
	} // switch state

    // Set a cooke that says they were anonymous.  Later, if another
    // request is sent with this cookie, we'll assume they saw the info
    // page.  We're only using a cookie because I don't want to have the
    // bother of defining magic urls to go to.
    if (sendcookie) {
      apr_table_add(r->headers_out, "Set-Cookie", cookieval);
    }

    if (decline) {
      return DECLINED;
    } // if we should send them to the real content handler

    // Send a header
    ap_set_content_type(r, "text/html");
    /*
     * If we're only supposed to send header information (HEAD request), we're
     * already there.
     */
    if (r->header_only) {
        return OK;
    }


    showfile = 0;
    use_default_text = 0;
    switch (state) {
    case ATO_STATE_ANON_IGNORANT:
      // Let's read the template file.
      if (apr_file_open(&template, our_config->anon_url, APR_READ, 1, r->pool))
      {
        // An error has occurred.
        use_default_text = 1;
        default_txt = apr_pstrcat(
            r->pool,
            DOCTYPE_XHTML_1_0S,
            "<html>\n",
            "<head>\n",
            "<title>Anonymous Request</title>\n",
            "</head>\n",
            "<body>\n",
            "<h1>Anonymous Request</h1>\n\n",
            "<p>You are attempting to use this page anonymously.  ",
            "However, this page is password-protected.</p>\n",
            "<p><a href=\"\">Click here</a> and you will be prompted for ",
            "your username and password.</p>\n",
            "</body>\n",
            "</html>\n",
            NULL);
      } // if we couldn't open the file
      showfile = 1;
      break;

    case ATO_STATE_NONANON_IGNORANT:
      if (apr_file_open(&template, our_config->nonanon_url, APR_READ, 1, r->pool)) {
        // An error has occurred
        use_default_text = 1;
        default_txt = apr_pstrcat(
            r->pool,
            DOCTYPE_XHTML_1_0S,
            "<html>\n",
            "<head>\n",
            "<title>Timed-Out Request</title>\n",
            "</head>\n",
            "<body>\n",
            "<h1>Timed-Out Request</h1>\n\n",
            "<p>You have successfully logged in to this page.  However, ",
            "it appears that the last time you logged in was long enough ago ",
            "that your session is expired.</p>\n",
            "<p>Therefore, for your security, you will be prompted for your ",
            "username and password again, to ensure that somebody is not ",
            "simply using a computer that you left unattended.</p>",
            "<p><a href=\"\">Click here</a> and you will be prompted for ",
            "your username and password.</p>\n",
            "</body>\n",
            "</html>\n",
            NULL);
      } // if we couldn't open the file
      showfile = 1;
      break;

    default:
      showfile = 0;
      break;
    } // switch state

    if (showfile) {
      if (!use_default_text) {
        if (apr_file_info_get(&info, APR_FINFO_SIZE, template)) {
            // An error has occurred.
        } // if we couldn't get the file size

        filesize = info.size;
        txt = apr_palloc(r->pool, filesize+1);
        if (apr_file_read(template, txt, &filesize)) {
            // An error has occurred
        } // if we couldn't read the file
        txt[filesize] = '\0';
        retval = txt;
      } // if we're actually reading from the file
      else {
        retval = default_txt;
      } // else we're using the default text

      /*
       * Now send our actual output.  Since we tagged this as being
       * "text/html", we need to embed any HTML.
       */
      ap_rputs(retval, r);

      return OK;
    } // if we're showing a file

    // I have nothing to say to them.
    return DECLINED;
} // static int x_handler

/*--------------------------------------------------------------------------*/
/*                                                                          */
/* Which functions are responsible for which hooks in the server.           */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*
 * Each function our module provides to handle a particular hook is
 * specified here.  The functions are registered using
 * ap_hook_foo(name, predecessors, successors, position)
 * where foo is the name of the hook.
 *
 * The args are as follows:
 * name         -> the name of the function to call.
 * predecessors -> a list of modules whose calls to this hook must be
 *                 invoked before this module.
 * successors   -> a list of modules whose calls to this hook must be
 *                 invoked after this module.
 * position     -> The relative position of this module.  One of
 *                 APR_HOOK_FIRST, APR_HOOK_MIDDLE, or APR_HOOK_LAST.
 *                 Most modules will use APR_HOOK_MIDDLE.  If multiple
 *                 modules use the same relative position, Apache will
 *                 determine which to call first.
 *                 If your module relies on another module to run first,
 *                 or another module running after yours, use the
 *                 predecessors and/or successors.
 *
 * The number in brackets indicates the order in which the routine is called
 * during request processing.  Note that not all routines are necessarily
 * called (such as if a resource doesn't have access restrictions).
 * The actual delivery of content to the browser [9] is not handled by
 * a hook; see the handler declarations below.
 */
static void x_register_hooks(apr_pool_t *p)
{
    ap_hook_post_config(x_post_config, NULL, NULL, APR_HOOK_MIDDLE);
    ap_hook_check_user_id(x_check_user_id, NULL, NULL, APR_HOOK_FIRST);
    ap_hook_fixups(x_fixer_upper, NULL, NULL, APR_HOOK_LAST);
    ap_hook_handler(x_handler, NULL, NULL, APR_HOOK_FIRST);
}


/*--------------------------------------------------------------------------*/
/*                                                                          */
/* Finally, the list of callback routines and data structures that provide  */
/* the static hooks into our module from the other parts of the server.     */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*
 * Module definition for configuration.  If a particular callback is not
 * needed, replace its routine name below with the word NULL.
 */

module AP_MODULE_DECLARE_DATA auth_timeout_module =
{
    STANDARD20_MODULE_STUFF,
    x_create_dir_config,    /* per-directory config creator */
    x_merge_config,     /* dir config merger */
    x_create_server_config, /* server config creator */
    x_merge_config,  /* server config merger */
    x_cmds,                 /* command table */
    x_register_hooks,       /* set up other request processing hooks */
};
