/*  Copyright 2008 Linux Box, Inc.  Ryan Hughes <ryan@linuxbox.com>.
 *  Licensed under the Apache License, Version 2.0 (the "License"); you may not
 *  use this file except in compliance with the License. You may obtain a copy
 *  of the License at 
 *          http://www.apache.org/licenses/LICENSE-2.0 
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *  License for the specific language governing permissions and limitations
 *  under the License.
 */

/**
 * Probe your dbm file, to see the current state.
 * Prints out all the keys and values in your database.
 */
#include "apr_base64.h"
#include "apr_dbm.h"
#include "apr_errno.h"
#include "apr_file_io.h"
#include "apr_md5.h"
#include "apr_shm.h"
#include "apr_strings.h"

#include <ctype.h>
#include <stdarg.h>
#include <stdio.h>
#include <time.h>

#define ERRSTR_SIZE 1000

void ERR(apr_pool_t *pool, char *fmt, ...)
{
  va_list ap;
  char *msg;
  apr_file_t *erfile;
  
  va_start(ap, fmt);
  msg = apr_pvsprintf(pool, fmt, ap);
  
  apr_file_open_stderr(&erfile, pool);
  apr_file_puts(msg, erfile);
  va_end(ap);
}

/**
 * Open the DBM file.
 * @param[out] dbm The dbm handle, once open.
 * @param filename The name of the dbm file.
 * @return 0 or error.
 */
static apr_status_t open_db(apr_dbm_t **dbm, char *filename, apr_pool_t *pool)
{
	apr_status_t err;
	char errstr[ERRSTR_SIZE];

	err=apr_dbm_open(
						dbm, filename, 
						APR_DBM_RWCREATE, APR_OS_DEFAULT, pool);
	if (err) {
		apr_strerror(err, errstr, ERRSTR_SIZE);
		ERR(pool, "Failed to open dbm file (%s).  Error: %s.",
				filename, 
				errstr);
		return err;
	} // if the open of the dbm file failed

	return APR_SUCCESS;
} // apr_status_t open_db

/**
 * Lock the database.
 * There are several non-atomic database operations that we do.  Therefore,
 * they must be locked.  We lock only the file that we're using for this
 * request.
 * @param filename The name of the dbm file.
 * @param pool The memory pool to use.
 * @param[out] db_lockfile The lockfile we created, for later unlocking.
 * @return 0 or error.
 */
static apr_status_t lock_db(char *filename, apr_pool_t *pool, apr_file_t **db_lockfile)
{
  char *lockfile_name;
	char errstr[ERRSTR_SIZE];
	apr_status_t err;

  lockfile_name = apr_pstrcat(pool, filename, ".lock", NULL);
  err = apr_file_open(
      db_lockfile, lockfile_name, 
      APR_CREATE|APR_WRITE|APR_BINARY|APR_DELONCLOSE, 0, pool);
  apr_file_perms_set(lockfile_name, APR_UREAD | APR_UWRITE | APR_GREAD | APR_GWRITE);

  if (err) {
		apr_strerror(err, errstr, ERRSTR_SIZE);
		ERR(pool, "Failed to create lock file (%s).  Error: %s.", 
				lockfile_name, 
				errstr);
		return err;
  } // if we couldn't open the file

  apr_file_lock(*db_lockfile, APR_FLOCK_EXCLUSIVE);

  return APR_SUCCESS;
} // apr_status_t lock_db

/**
 * Unlock the databases.
 * We only lock/unlock the actual db used for this request.
 * @param db_lockfile The lock file.
 * @return 0.
 */
static apr_status_t unlock_db(apr_file_t *db_lockfile)
{
  apr_file_unlock(db_lockfile);
  apr_file_close(db_lockfile);

  return APR_SUCCESS;
} // apr_status_t unlock_db


static void probe_dbm(char *filename, apr_pool_t *pool)
{
  int howmany;
  int stopnow;

  apr_datum_t key_datum;
  apr_datum_t val_datum;

  apr_file_t *db_lockfile;
	char errstr[ERRSTR_SIZE];

  int last_access_time;
	apr_status_t err;
	apr_file_t *out;

	apr_file_open_stdout(&out, pool);

  apr_dbm_t *sessions;

  open_db(&sessions, filename, pool);

  // Enter critical section
  if ((err = lock_db(filename, pool, &db_lockfile))) {
		apr_strerror(err, errstr, ERRSTR_SIZE);
		ERR(pool, "Failed to open lock file.  Error: %s.", errstr);
  } // if the locking failed

  howmany=0;
  for (stopnow=apr_dbm_firstkey(sessions, &key_datum);
        !stopnow && key_datum.dsize;
        stopnow=apr_dbm_nextkey(sessions, &key_datum))
  {
    howmany += 1;
  } // for all the keys

  apr_file_printf(out, "There are %d entries\n", howmany);

  for (stopnow=apr_dbm_firstkey(sessions, &key_datum);
        !stopnow && key_datum.dsize;
        stopnow=apr_dbm_nextkey(sessions, &key_datum))
  {
      apr_dbm_fetch(sessions, key_datum, &val_datum);
      memcpy(&last_access_time, val_datum.dptr, sizeof(int));

      apr_file_printf(out, "%s: %d\n", key_datum.dptr, last_access_time);
  } // for all the things in the hash table

  // End critical section
  unlock_db(db_lockfile);
} // void print_what_we_have

int main(int argc, char **argv) {
  apr_pool_t *mainpool;

  apr_pool_initialize();
  apr_pool_create(&mainpool, NULL);

  if (argc > 1) {
    probe_dbm(argv[1], mainpool);
  } else {
    ERR(mainpool, "Usage: %s <dbm-file>", argv[0]);
  }

	return 0;
}
